/*
 * Copyright (c) 2013 Peter S. May
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 * associated documentation files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 * NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

package us.wxy.iomux;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.util.concurrent.Semaphore;

/**
 * Wraps a {@link ByteInput} with a blocking mechanism accessible through a {@link ByteIOInput}.
 * <p>
 * This class doesn't involve itself with the process by which the data is produced for the
 * non-blocking input; instead, every time data is added, {@link #attention} must be called. Calling
 * this method unblocks one of the waiting read threads, which attempts a read. If the input is open
 * and the read attempt succeeds, the read will unblock another thread, and so forth. If
 * {@link #isClosed} returns {@code true} or if the read attempt returns empty, the thread returns
 * to waiting and no other thread is unblocked.
 * <p>
 * Subclasses of this class can override {@link #isClosed} to allow the producer that feeds the
 * input to tell the blocking mechanism that there is no more data to block for. If this method is
 * overridden, the producer should also call {@link #attentionAll()}, which unblocks all detected
 * blocking reads, allowing them to determine that the input is closed and return.
 * <p>
 * Spurious attention signals should not negatively impact the correctness of the behavior, but
 * excessive unnecessary calls could potentially affect efficiency.
 */
public class BlockingByteInput implements ByteIOInput {
  private static abstract class Getter {
    public abstract int get() throws IOException;
  }

  // Shorthand for adding an InterruptedException to a new InterruptedIOException.
  private static InterruptedIOException withCause(InterruptedIOException z, InterruptedException e) {
    z.initCause(e);
    return z;
  }

  private final ByteInput input;

  private final Semaphore s = new Semaphore(0, true);

  public BlockingByteInput(ByteInput input) {
    if (input == null) throw new NullPointerException();
    this.input = input;
  }

  /**
   * Called by the actual source of data whenever new data might be available.
   */
  public void attention() {
    s.release();
  }

  /**
   * Called by the actual source of data when it has changed the state of {@link #isClosed()} to
   * {@code true}.
   */
  public void attentionAll() {
    while (s.hasQueuedThreads())
      s.release();
  }

  private int blockingGet(Getter g) throws IOException {
    int len = 0;
    boolean isCurrentlyOpen = true;
    try {
      // This acquires a signaled availability, but discards it
      // (not to be regenerated) if no data is actually available.
      // If data was available for the read, or if an interrupt
      // occurs, the acquired signal is released to be acquired
      // elsewhere.
      while ((isCurrentlyOpen = !isClosed()) && ((len = g.get()) <= 0))
        s.acquire();
      return len;
    } catch (InterruptedException e) {
      throw withCause(new InterruptedIOException(), e);
    } finally {
      if (isCurrentlyOpen) s.release();
    }
  }

  /**
   * Called by the blocking wrapper to determine whether the input source feeding this object is
   * closed. If this method returns {@code true}, a new read that would normally block read will
   * return empty instead. The default implementation always returns {@code false}.
   * <p>
   * If changed in a subclass to return {@code true}, the closing action should immediately
   * afterward call {@link #attentionAll} to unblock waiting readers, allowing them to discover that
   * the object is now closed.
   * 
   * @return Whether the input source is closed
   */
  public boolean isClosed() {
    return false;
  }

  @Override
  public int readBytes(final byte[] dest, final int offset, final int length) throws IOException {
    return blockingGet(new Getter() {
      @Override
      public int get() {
        return input.getBytes(dest, offset, length);
      }
    });
  }

  @Override
  public int readBytes(final ByteArrayOutput dest) throws IOException {
    return blockingGet(new Getter() {
      @Override
      public int get() {
        return input.getBytes(dest);
      }
    });
  }

  @Override
  public int readBytes(final ByteArrayOutput dest, final int length) throws IOException {
    return blockingGet(new Getter() {
      @Override
      public int get() {
        return input.getBytes(dest, length);
      }
    });
  }

  @Override
  public int readBytes(final OutputStream dest) throws IOException {
    return blockingGet(new Getter() {
      @Override
      public int get() throws IOException {
        return input.getBytes(dest);
      }
    });
  }

  @Override
  public int readBytes(final OutputStream dest, final int length) throws IOException {
    return blockingGet(new Getter() {
      @Override
      public int get() throws IOException {
        return input.getBytes(dest, length);
      }
    });
  }
}
