/*
 * Copyright (c) 2013 Peter S. May
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 * associated documentation files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 * NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

package us.wxy.iomux;

/**
 * Plain interface for objects from which an array of byte data can be read.
 */
public interface ByteArrayInput {
  /**
   * Reads up to {@code length} bytes of data from this object into an array of bytes.
   * <p>
   * The interesting parts of the contract are these:
   * <ul>
   * <li>{@code dest} must be non-null.</li>
   * <li>{@code offset} must be nonnegative, {@code length} must be nonnegative, and their sum must
   * be less than or equal to {@code dest.length}. Validating these, if necessary, is the caller's
   * job; the method may also validate them or opt not to do so.
   * <ul>
   * <li>If {@code length} is 0, the method may exit early, returning 0, without first validating
   * {@code dest} or {@code offset}.</li>
   * <li>The method may treat a negative length as 0 instead of throwing.</li>
   * </ul>
   * </li>
   * <li>The method may copy {@code length} bytes or fewer (possibly even 0) bytes. Except where
   * {@code length} is 0, a read of zero length should not occur unless there is no data readily
   * available (e.g. without blocking), but the definition of availability depends on the
   * implementation.</li>
   * <li>The method must not return a negative number. If no bytes are copied, regardless of reason,
   * the return value is 0. There is no return value that distinguishes an EOF-like condition from
   * another zero-length read; this sort of signaling must be handled out-of-band.</li>
   * <li>Unless otherwise documented, the method does not block to wait for more input; it reports
   * only on the input that is already immediately available.</li>
   * </ul>
   * 
   * @param dest The destination buffer for the data
   * @param offset The start index in {@code dest} to begin copying
   * @param length The maximum number of bytes to copy
   * @return The number of bytes actually copied
   */
  public int getBytes(byte[] dest, int offset, int length);
}
