/*
 * Copyright (c) 2013 Peter S. May
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 * associated documentation files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 * NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

package us.wxy.iomux;

/**
 * Plain interface for objects to which an array of byte data can be read.
 */
public interface ByteArrayOutput {
  /**
   * Writes up to {@code length} bytes of data to this object from an array of bytes.
   * <p>
   * The interesting parts of the contract are these:
   * <ul>
   * <li>{@code src} must be non-null.</li>
   * <li>{@code offset} must be nonnegative, {@code length} must be nonnegative, and their sum must
   * be less than or equal to {@code src.length}. Validating these, if necessary, is the caller's
   * job; the method may also validate them or opt not to do so.
   * <ul>
   * <li>If {@code length} is 0, the method may exit early, returning 0, without first validating
   * {@code src} or {@code offset}.</li>
   * <li>The method may treat a negative length as 0 instead of throwing.</li>
   * </ul>
   * </li>
   * <li>The method may copy {@code length} bytes or fewer (possibly even 0) bytes. Except where
   * {@code length} is 0, a write of zero length should not occur unless the object is exhausted
   * (i.e., has no writable space available), where the definition of exhaustion depends on the
   * implementation. A zero-length write does not necessarily imply that the object's exhaustion is
   * permanent, but might be treated as a cue for the caller to check for an out-of-band signal.</li>
   * <li>The method must not return a negative number. If no bytes are copied, regardless of reason,
   * the return value is 0.</li>
   * <li>Unless otherwise documented, the method does not block; it receives as much input as it can
   * immediately process and then returns.</li>
   * </ul>
   * 
   * @param src The source buffer for the data
   * @param offset The start index in {@code src} to begin copying
   * @param length The maximum number of bytes to copy
   * @return The number of bytes actually copied
   */
  public int putBytes(byte[] src, int offset, int length);
}
