/*
 * Copyright (c) 2013 Peter S. May
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 * associated documentation files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 * NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

package us.wxy.iomux;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.OutputStream;


/**
 * An analogue to {@link ByteInput}, but with the added implication that a call requesting a length
 * of greater than zero will block until at least one byte is available for read.
 * <p>
 * Each overload of {@code readBytes} returns the number of bytes copied. Unless the requested
 * length is 0, the return value should never be 0 unless the end of data has been permanently
 * reached (similar to the meaning of -1 when reading an input stream).
 * <p>
 * All methods can throw {@link IOException}. Blocking behavior is likely to be implemented in terms
 * of either some other blocking I/O primitive (such as {@link java.io.InputStream}) or with some
 * wait/notify condition that throws {@link InterruptedException}. For the latter, the exception
 * should be caught, and a {@link InterruptedIOException} should be thrown in its place. (
 * {@link java.io.PipedInputStream}, for example, does this.)
 */
public interface ByteIOInput {

  /**
   * Reads data out of this object into an array.
   * 
   * @param dest An array to copy data into
   * @param offset The start index of {@code dest} to begin copying
   * @param length The maximum number of bytes to attempt to copy
   * @return The number of bytes actually copied, which may be {@code length} or less, if less data
   *         is available, or 0 only if either {@code length} is 0 or no more data can be read
   * @throws IOException If an I/O error occurs while waiting for data
   */
  public abstract int readBytes(byte[] dest, int offset, int length) throws IOException;

  /**
   * Reads data out of this object into a {@link ByteArrayOutput}. This is equivalent to
   * {@code readBytes(dest, Integer.MAX_VALUE)}.
   * 
   * @param dest An {@link ByteArrayOutput} to copy data into
   * @return The number of bytes actually copied, which is 0 only if no more data can be read
   * @throws IOException If an I/O error occurs while waiting for data
   */
  public abstract int readBytes(ByteArrayOutput dest) throws IOException;

  /**
   * Reads data out of this object into a {@link ByteArrayOutput}.
   * 
   * @param dest An {@link ByteArrayOutput} to copy data into
   * @param length The maximum number of bytes to attempt to copy
   * @return The number of bytes actually copied, which may be {@code length} or less, if less data
   *         is available, or 0 only if either {@code length} is 0 or no more data can be read
   * @throws IOException If an I/O error occurs while waiting for data
   */
  public abstract int readBytes(ByteArrayOutput dest, int length) throws IOException;

  /**
   * Reads data out of this object into an output stream. This is equivalent to
   * {@code readBytes(dest, Integer.MAX_VALUE)}.
   * 
   * @param dest An {@link OutputStream} to copy data into
   * @return The number of bytes actually copied, which is 0 only if no more data can be read
   * @throws IOException If an I/O error occurs while waiting for data or writing to {@code dest}
   */
  public abstract int readBytes(OutputStream dest) throws IOException;

  /**
   * Reads data out of this object into an output stream.
   * 
   * @param dest An {@link OutputStream} to copy data into
   * @param length The maximum number of bytes to attempt to copy
   * @return The number of bytes actually copied, which may be {@code length} or less, if less data
   *         is available, or 0 only if either {@code length} is 0 or no more data can be read
   * @throws IOException If an I/O error occurs while waiting for data or writing to {@code dest}
   */
  public abstract int readBytes(OutputStream dest, int length) throws IOException;
}
