/*
 * Copyright (c) 2013 Peter S. May
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 * associated documentation files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 * NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

package us.wxy.iomux;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;

/**
 * A single-use, non-circular, non-blocking FIFO implemented on top of a byte array. Each byte is
 * written once, then read once, and then is not used again. After writing or reading the last byte
 * in the array, any further attempts to write or read, respectively, will result in a zero-length
 * no-op.
 * <p>
 * The buffer is not synchronized, but it is intended to be thread-safe as long as at most one
 * thread is reading and at most one thread is writing. Any write operations are performed before
 * the write location is incremented, and the write location is set {@code volatile} to encourage
 * the correct sequence.
 * <p>
 * If the length of an operation would, due either to a specified length or to available space, be
 * zero or negative, the operation is silently skipped and 0 is returned.
 * <p>
 * A reset mechanism is omitted intentionally. If pooling is necessary, pool the array itself and
 * create a new FIFO for each time that array is reused.
 */
public class FixedByteArrayFifo implements ByteInput, ByteOutput {
  private static int min(int a, int b) {
    return (a < b) ? a : b;
  }

  private final byte[] block;
  private int readLoc = 0;
  private volatile int writeLoc = 0;

  /**
   * Creates a FIFO with the given byte array.
   * 
   * Note: This constructor does not guard against null or zero-length arrays; the calling code is
   * presumed to know better.
   */
  public FixedByteArrayFifo(byte[] block) {
    this.block = block;
  }

  // Write from ByteInput, length already processed
  private int copyFromByteInput(ByteArrayInput src, int offset, int length) {
    if (length <= 0) return 0;
    int rlen = src.getBytes(block, offset, length);
    writeLoc = offset + ((rlen <= 0) ? 0 : rlen);
    return rlen;
  }

  // Write from InputStream, length already processed
  private int copyFromStream(InputStream src, int offset, int length) throws IOException {
    if (length <= 0) return 0;
    int rlen = src.read(block, offset, length);
    writeLoc = offset + ((rlen <= 0) ? 0 : rlen);
    return rlen;
  }

  // Read to ByteOutput, length already processed
  private int copyToByteOutput(ByteArrayOutput dest, int length) {
    if (length <= 0) return 0;
    int wlen = dest.putBytes(block, readLoc, length);
    readLoc += wlen;
    return wlen;
  }

  // Read to OutputStream, length already processed
  private int copyToStream(OutputStream dest, int length) throws IOException {
    if (length <= 0) return 0;
    dest.write(block, readLoc, length);
    readLoc += length;
    return length;
  }

  // Read to PrintStream, length already processed
  private int copyToStream(PrintStream dest, int length) {
    if (length <= 0) return 0;
    dest.write(block, readLoc, length);
    readLoc += length;
    return length;
  }

  @Override
  public int getBytes(byte[] dest, int offset, int length) {
    final int len = min(length, writeLoc - readLoc);
    if (len <= 0) return 0;
    System.arraycopy(block, readLoc, dest, offset, len);
    readLoc += len;
    return len;
  }

  @Override
  public int getBytes(ByteArrayOutput dest) {
    return copyToByteOutput(dest, writeLoc - readLoc);
  }

  @Override
  public int getBytes(ByteArrayOutput dest, int length) {
    return copyToByteOutput(dest, min(length, writeLoc - readLoc));
  }

  @Override
  public int getBytes(OutputStream dest) throws IOException {
    return copyToStream(dest, writeLoc - readLoc);
  }

  @Override
  public int getBytes(OutputStream dest, int length) throws IOException {
    return copyToStream(dest, min(length, writeLoc - readLoc));
  }

  @Override
  public int getBytes(PrintStream dest) {
    return copyToStream(dest, writeLoc - readLoc);
  }

  @Override
  public int getBytes(PrintStream dest, int length) {
    return copyToStream(dest, min(length, writeLoc - readLoc));
  }

  @Override
  public int putBytes(byte[] src, int offset, int length) {
    final int off = writeLoc;
    final int len = min(length, block.length - off);
    if (len <= 0) return 0;
    System.arraycopy(src, offset, block, off, len);
    writeLoc = off + len;
    return len;
  }

  @Override
  public int putBytes(ByteArrayInput src) {
    final int off = writeLoc;
    return copyFromByteInput(src, off, block.length - off);
  }

  @Override
  public int putBytes(ByteArrayInput src, int length) {
    final int off = writeLoc;
    return copyFromByteInput(src, off, min(length, block.length - off));
  }

  @Override
  public int putBytes(InputStream src) throws IOException {
    final int off = writeLoc;
    return copyFromStream(src, off, block.length - off);
  }

  @Override
  public int putBytes(InputStream src, int length) throws IOException {
    final int off = writeLoc;
    return copyFromStream(src, off, min(length, block.length - off));
  }


  /**
   * Returns whether the buffer is exhausted for reads and any further read on this buffer will be
   * zero-length.
   * 
   * @return {@code true} if all readable data has been exhausted and no further data can be
   *         written; {@code false} if there is still unread data or more data can be written.
   */
  public boolean readEnd() {
    return readLoc >= block.length;
  }

  /**
   * Returns whether the buffer is exhausted for writes and any further write on this buffer will be
   * zero-length.
   * 
   * @return {@code true} if no further data can be written; {@code false} if more data can still be
   *         written.
   */
  public boolean writeEnd() {
    return writeLoc >= block.length;
  }

  /**
   * Returns number of bytes available for read.
   * 
   * @return Number of bytes available for read
   */
  public int availableForRead() {
    return writeLoc - readLoc;
  }

  /**
   * Returns number of bytes available for write.
   * 
   * @return Number of bytes available for write
   */
  public int availableForWrite() {
    return block.length - writeLoc;
  }

  /**
   * A fifo with zero space, with all operations hardcoded as zero-length no-ops.
   */
  public static final FixedByteArrayFifo EMPTY = new FixedByteArrayFifo(null) {

    @Override
    public int getBytes(byte[] dest, int offset, int length) {
      return 0;
    }

    @Override
    public int getBytes(ByteArrayOutput dest) {
      return 0;
    }

    @Override
    public int getBytes(ByteArrayOutput dest, int length) {
      return 0;
    }

    @Override
    public int getBytes(OutputStream dest) throws IOException {
      return 0;
    }

    @Override
    public int getBytes(OutputStream dest, int length) throws IOException {
      return 0;
    }

    @Override
    public int getBytes(PrintStream dest) {
      return 0;
    }

    @Override
    public int getBytes(PrintStream dest, int length) {
      return 0;
    }

    @Override
    public int putBytes(byte[] src, int offset, int length) {
      return 0;
    }

    @Override
    public int putBytes(ByteArrayInput src) {
      return 0;
    }

    @Override
    public int putBytes(ByteArrayInput src, int length) {
      return 0;
    }

    @Override
    public int putBytes(InputStream src) throws IOException {
      return 0;
    }

    @Override
    public int putBytes(InputStream src, int length) throws IOException {
      return 0;
    }

    @Override
    public boolean readEnd() {
      return true;
    }

    @Override
    public boolean writeEnd() {
      return true;
    }

    @Override
    public int availableForRead() {
      return 0;
    }

    @Override
    public int availableForWrite() {
      return 0;
    }
  };

}
