/*
 * Copyright (c) 2013 Peter S. May
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 * associated documentation files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 * NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

package us.wxy.iomux;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.concurrent.ConcurrentLinkedQueue;

/**
 * A non-blocking but stream-like FIFO that automatically expands and contracts based on the size of
 * its contents.
 * <p>
 * The buffer is not synchronized, but it is intended to be thread-safe as long as at most one
 * thread is reading and at most one thread is writing.
 * <p>
 * A read or write operation is allowed to act on arbitrarily fewer bytes than specified in the
 * length argument.
 * <ul>
 * <li>A write operation from an array, for a number of bytes greater than zero, will write at least
 * one byte.</li>
 * <li>A write operation from a stream, for a number of bytes greater than zero, will write at least
 * one byte unless no data can be read from the stream.</li>
 * <li>A read operation, for a number of bytes greater than zero, will read at least one byte unless
 * the buffer is empty.</li>
 * <li>Attempting to read <var>n</var> bytes may result in a read of fewer than <var>n</var> bytes,
 * even if there are <var>n</var> or more bytes available in the buffer.</li>
 * <li>Attempting to write <var>n</var> bytes may result in a write of fewer than <var>n</var>
 * bytes, even if it is possible to write more.</li>
 * </ul>
 * <p>
 * A read or write may need to be called repeatedly to fill a request.
 * 
 * <pre>
 * // An example of repeated call needed for write
 * for(int offset = 0, len; offset &lt; data.length; offset += len) {
 *   len = fifo.write(data, offset, data.length - offset);
 * }
 * </pre>
 * <p>
 * There is no concept of EOF here; the buffer is perpetual and uncloseable. A read of 0 bytes (for
 * a nonzero request) indicates the end only of <em>currently available</em> data. Once a non-empty
 * write has occurred, a non-empty read is once again possible.
 * <p>
 * Implementation notes:
 * <p>
 * This implementation is based on a queue of block-sized arrays. When more write space is needed, a
 * new array is allocated; after an array has been written and read, it is discarded. No pooling
 * takes place.
 * <p>
 * Any read or write method available here works with exactly one of the arrays. Depending on the
 * arrangement of data, a read or write may need to be executed in multiple passes.
 * <p>
 */
public class FlexByteFifo implements ByteInput, ByteOutput {
  private ConcurrentLinkedQueue<FixedByteArrayFifo> blocks =
      new ConcurrentLinkedQueue<FixedByteArrayFifo>();

  private static final int BLOCK_SIZE = 8192;

  private FixedByteArrayFifo writeBlock = FixedByteArrayFifo.EMPTY;

  public FlexByteFifo() {}

  // Get the next block that will be readable, or a dummy block if there is none.
  private FixedByteArrayFifo getReadBlock() {
    FixedByteArrayFifo block = blocks.peek();
    return (block == null) ? FixedByteArrayFifo.EMPTY : block;
  }

  // Skip past blocks that can no longer be read.
  private void advanceReadBlock() {
    FixedByteArrayFifo block;
    while (((block = blocks.peek()) != null) && block.readEnd()) {
      blocks.remove();
    }
  }

  // Get the block currently being used for writes, generating a new one if
  // there is no space remaining in the current one.
  private FixedByteArrayFifo getWriteBlock() {
    if (writeBlock.writeEnd()) {
      writeBlock = new FixedByteArrayFifo(new byte[BLOCK_SIZE]);
      // The block is added to the queue immediately so that it can be
      // read before it is finished being written
      blocks.add(writeBlock);
    }
    return writeBlock;
  }

  @Override
  public int getBytes(byte[] dest, int offset, int length) {
    int r = getReadBlock().getBytes(dest, offset, length);
    advanceReadBlock();
    return r;
  }

  @Override
  public int getBytes(OutputStream dest, int length) throws IOException {
    int r = getReadBlock().getBytes(dest, length);
    advanceReadBlock();
    return r;
  }

  @Override
  public int getBytes(OutputStream dest) throws IOException {
    int r = getReadBlock().getBytes(dest);
    advanceReadBlock();
    return r;
  }

  @Override
  public int getBytes(PrintStream dest, int length) {
    int r = getReadBlock().getBytes(dest, length);
    advanceReadBlock();
    return r;
  }

  @Override
  public int getBytes(PrintStream dest) {
    int r = getReadBlock().getBytes(dest);
    advanceReadBlock();
    return r;
  }

  @Override
  public int getBytes(ByteArrayOutput dest, int length) {
    int r = getReadBlock().getBytes(dest, length);
    advanceReadBlock();
    return r;
  }

  @Override
  public int getBytes(ByteArrayOutput dest) {
    int r = getReadBlock().getBytes(dest);
    advanceReadBlock();
    return r;
  }

  @Override
  public int putBytes(byte[] src, int offset, int length) {
    if (length <= 0) return 0;
    return getWriteBlock().putBytes(src, offset, length);
  }

  @Override
  public int putBytes(InputStream src, int length) throws IOException {
    if (length <= 0) return 0;
    return getWriteBlock().putBytes(src, length);
  }

  @Override
  public int putBytes(InputStream src) throws IOException {
    return getWriteBlock().putBytes(src);
  }

  @Override
  public int putBytes(ByteArrayInput src, int length) {
    if (length <= 0) return 0;
    return getWriteBlock().putBytes(src, length);
  }

  @Override
  public int putBytes(ByteArrayInput src) {
    return getWriteBlock().putBytes(src);
  }

  /**
   * Returns number of bytes immediately available for read.
   * 
   * @return Number of bytes available for read before another write takes place
   */
  public int availableForRead() {
    return getReadBlock().availableForRead();
  }

  /**
   * Returns number of bytes immediately available for write.
   * 
   * Implementation note: This method will never return 0; if the current block is exhausted and the
   * next block has not yet been allocated, the size of the next block that would be allocated is
   * returned.
   * 
   * @return Number of bytes available for write before another block is allocated.
   */
  public int availableForWrite() {
    int len = (writeBlock == null) ? 0 : writeBlock.availableForWrite();
    return (len <= 0) ? BLOCK_SIZE : len;
  }
}
