/*
 * Copyright (c) 2013 Peter S. May
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 * associated documentation files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 * NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

package us.wxy.iomux;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;

/**
 * Wraps an existing {@link ByteInput}, adding repetitions, so that the read length is only ever
 * less than the requested length if the input runs out of immediately available data.
 * <p>
 * A {@code ByteInput} itself may copy arbitrarily less than the requested length for any reason,
 * even if more data is available without waiting; this wrapper repeats such an operation until the
 * requested length is reached or a partial read length is 0, which indicates that no more
 * information is immediately available.
 */
public class RepeatingByteInput implements ByteInput {
  private ByteInput input;

  /**
   * Wraps this existing byte input.
   * 
   * @param input An {@link ByteInput} to be wrapped
   */
  public RepeatingByteInput(ByteInput input) {
    if (input == null) throw new NullPointerException();
    this.input = input;
  }

  /**
   * {@inheritDoc}
   * 
   * This implementation calls the wrapped input repeatedly until either the full requested length
   * has been read or a zero-length read occurs (indicating that no more data is immediately
   * available).
   */
  @Override
  public int getBytes(byte[] dest, int offset, int length) {
    int total = 0;
    while (total < length) {
      int len = input.getBytes(dest, offset + total, length - total);
      if (len <= 0) break;
      total += len;
    }
    return total;
  }

  /**
   * {@inheritDoc}
   * 
   * This implementation calls the wrapped input repeatedly until a zero-length read occurs
   * (indicating that no more data is immediately available).
   */
  @Override
  public int getBytes(ByteArrayOutput dest) {
    int total = 0;
    for (;;) {
      int len = input.getBytes(dest);
      if (len <= 0) return total;
      total += len;
    }
  }

  /**
   * {@inheritDoc}
   * 
   * This implementation calls the wrapped input repeatedly until either the full requested length
   * has been read or a zero-length read occurs (indicating that no more data is immediately
   * available).
   */
  @Override
  public int getBytes(ByteArrayOutput dest, int length) {
    int total = 0;
    while (total < length) {
      int len = input.getBytes(dest, length - total);
      if (len <= 0) break;
      total += len;
    }
    return total;
  }

  /**
   * {@inheritDoc}
   * 
   * This implementation calls the wrapped input repeatedly until a zero-length read occurs
   * (indicating that no more data is immediately available).
   */
  @Override
  public int getBytes(OutputStream dest) throws IOException {
    int total = 0;
    for (;;) {
      int len = input.getBytes(dest);
      if (len <= 0) return total;
      total += len;
    }
  }

  /**
   * {@inheritDoc}
   * 
   * This implementation calls the wrapped input repeatedly until either the full requested length
   * has been read or a zero-length read occurs (indicating that no more data is immediately
   * available).
   */
  @Override
  public int getBytes(OutputStream dest, int length) throws IOException {
    int total = 0;
    while (total < length) {
      int len = input.getBytes(dest, length - total);
      if (len <= 0) break;
      total += len;
    }
    return total;
  }

  /**
   * {@inheritDoc}
   * 
   * This implementation calls the wrapped input repeatedly until a zero-length read occurs
   * (indicating that no more data is immediately available).
   */
  @Override
  public int getBytes(PrintStream dest) {
    int total = 0;
    for (;;) {
      int len = input.getBytes(dest);
      if (len <= 0) return total;
      total += len;
    }
  }

  /**
   * {@inheritDoc}
   * 
   * This implementation calls the wrapped input repeatedly until either the full requested length
   * has been read or a zero-length read occurs (indicating that no more data is immediately
   * available).
   */
  @Override
  public int getBytes(PrintStream dest, int length) {
    int total = 0;
    while (total < length) {
      int len = input.getBytes(dest, length - total);
      if (len <= 0) break;
      total += len;
    }
    return total;
  }
}
