/*
 * Copyright (c) 2013 Peter S. May
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 * associated documentation files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 * NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

package us.wxy.iomux;

import java.io.IOException;
import java.io.InputStream;

/**
 * Wraps an existing {@link ByteOutput}, adding repetitions, so that the write length is only ever
 * less than the requested length if the output runs out of immediately available space.
 * <p>
 * A {@code ByteOutput} itself may copy arbitrarily less than the requested length for any reason,
 * even if more space is available without waiting; this wrapper repeats such an operation until the
 * requested length is reached or a partial write length is 0, which indicates that no more space is
 * immediately available.
 */
public class RepeatingByteOutput implements ByteOutput {
  private ByteOutput output;

  /**
   * Wraps this existing byte output.
   * 
   * @param output An {@link ByteOutput} to be wrapped
   */
  public RepeatingByteOutput(ByteOutput output) {
    if (output == null) throw new NullPointerException();
    this.output = output;
  }

  /**
   * {@inheritDoc}
   * 
   * This implementation calls the wrapped output repeatedly until either the full requested length
   * has been written or a zero-length write occurs (indicating that no more space is immediately
   * available).
   */
  @Override
  public int putBytes(byte[] src, int offset, int length) {
    int total = 0;
    while (total < length) {
      int len = output.putBytes(src, offset + total, length - total);
      if (len <= 0) break;
      total += len;
    }
    return total;
  }

  /**
   * {@inheritDoc}
   * 
   * This implementation calls the wrapped output repeatedly until a zero-length write occurs
   * (indicating that no more space is immediately available).
   */
  @Override
  public int putBytes(ByteArrayInput src) {
    int total = 0;
    for (;;) {
      int len = output.putBytes(src);
      if (len <= 0) break;
      total += len;
    }
    return total;
  }

  /**
   * {@inheritDoc}
   * 
   * This implementation calls the wrapped output repeatedly until either the full requested length
   * has been written or a zero-length write occurs (indicating that no more space is immediately
   * available).
   */
  @Override
  public int putBytes(ByteArrayInput src, int length) {
    int total = 0;
    while (total < length) {
      int len = output.putBytes(src, length - total);
      if (len <= 0) break;
      total += len;
    }
    return total;
  }

  /**
   * {@inheritDoc}
   * 
   * This implementation calls the wrapped output repeatedly until a zero-length write occurs
   * (indicating that no more space is immediately available).
   */
  @Override
  public int putBytes(InputStream src) throws IOException {
    int total = 0;
    for (;;) {
      int len = output.putBytes(src);
      if (len <= 0) break;
      total += len;
    }
    return total;
  }

  /**
   * {@inheritDoc}
   * 
   * This implementation calls the wrapped output repeatedly until either the full requested length
   * has been written or a zero-length write occurs (indicating that no more space is immediately
   * available).
   */
  @Override
  public int putBytes(InputStream src, int length) throws IOException {
    int total = 0;
    while (total < length) {
      int len = output.putBytes(src, length - total);
      if (len <= 0) break;
      total += len;
    }
    return total;
  }
}
